/*
 *  SXAsmCore.c
 *  llasm
 *
 *  Created by Britt Durbrow on 7/1/06.
 *  Copyright 2006 Rattlesnake Hill Softworks. All rights reserved.
 *
 */

#include "SXAsmCore.h"

#include <stdio.h>
#include <strings.h>

char *formStrings[]=
{
	"",
	"#",
	"reg",
	"w",
	"reg, w",
	"w, reg",
	"w, #",
	"reg.bit"
};

opcode opcodes[]=
{
	{"add",FR_W,		0x1e0,0x00,0x1f,0x00,0x00},
	{"add",W_FR,		0x1c0,0x00,0x1f,0x00,0x00},
	
	{"clr",FR,			0x060,0x00,0x1f,0x00,0x00},
	{"clr",W,			0x400,0x00,0x00,0x00,0x00},
	{"clrwdt",FR,		0x004,0x00,0x00,0x00,0x00},
	
	{"dec",FR,			0x0e0,0x00,0x1f,0x00,0x00},
	{"dec",W_FR,		0x0c0,0x00,0x1f,0x00,0x00},
	{"decsz",FR,		0x2e0,0x00,0x1f,0x00,0x00},
	{"decsz",W_FR,		0x2c0,0x00,0x1f,0x00,0x00},

	{"inc",FR,			0x0a0,0x00,0x1f,0x00,0x00},
	{"inc",W_FR,		0x280,0x00,0x1f,0x00,0x00},
	{"incsz",FR,		0x3e0,0x00,0x1f,0x00,0x00},
	{"incsz",W_FR,		0x3c0,0x00,0x1f,0x00,0x00},

	{"rl",FR,			0x360,0x00,0x1f,0x00,0x00},
	{"rl",W_FR,			0x340,0x00,0x1f,0x00,0x00},
	
	{"rr",FR,			0x320,0x00,0x1f,0x00,0x00},
	{"rr",W_FR,			0x300,0x00,0x1f,0x00,0x00},

	{"sub",FR_W,		0x0a0,0x00,0x1f,0x00,0x00},
	{"sub",W_FR,		0x080,0x00,0x1f,0x00,0x00},
	
	{"and",FR_W,		0x160,0x00,0x1f,0x00,0x00},
	{"and",W_FR,		0x140,0x00,0x1f,0x00,0x00},
	{"and",W_L,			0xe00,0x00,0x00,0xff,0x00},
	
	{"or",FR_W,			0x120,0x00,0x1f,0x00,0x00},
	{"or",W_FR,			0x100,0x00,0x1f,0x00,0x00},
	{"or",W_L,			0xd00,0x00,0x00,0xff,0x00},
	
	{"xor",FR_W,		0x1a0,0x00,0x1f,0x00,0x00},
	{"xor",W_FR,		0x180,0x00,0x1f,0x00,0x00},
	{"xor",W_L,			0xf00,0x00,0x00,0xff,0x00},
	
	{"not",FR,			0x260,0x00,0x1f,0x00,0x00},
	{"not",W_FR,		0x240,0x00,0x1f,0x00,0x00},

	{"test",FR,			0x220,0x00,0x1f,0x00,0x00},
	{"test",W,			0xd00,0x00,0x1f,0x00,0x00},

	{"clrb",FR_B,		0x400,0xe0,0x1f,0x00,0x00},
	{"setb",FR_B,		0x500,0xe0,0x1f,0x00,0x00},
	
	{"clc",NO_OPS,		0x403,0x00,0x00,0x00,0x00},
	{"stc",NO_OPS,		0x503,0x00,0x00,0x00,0x00},
	
	{"sb",FR_B,			0x700,0xe0,0x1f,0x00,0x00},
	{"snb",FR_B,		0x600,0xe0,0x1f,0x00,0x00},
		
	{"mov",FR_W,		0x020,0x00,0x1f,0x00,0x00},
	{"mov",W_FR,		0x200,0x00,0x1f,0x00,0x00},
	{"mov",W_L,			0xc00,0x00,0x00,0xff,0x00},
		
	{"swap",FR,			0x3a0,0x00,0x1f,0x00,0x00},
	{"swap",W_FR,		0x380,0x00,0x1f,0x00,0x00},
	
	{"sc",NO_OPS,		0x703,0x00,0x00,0x00,0x00},
	{"snc",NO_OPS,		0x603,0x00,0x00,0x00,0x00},
	
	{"sz",NO_OPS,		0x743,0x00,0x00,0x00,0x00},
	{"snz",NO_OPS,		0x643,0x00,0x00,0x00,0x00},
	
	{"call",L,			0x900,0x00,0x00,0xff,0x00},
	{"jmp",L,			0xa00,0x00,0x00,0x1ff,0x00},
	{"ret",NO_OPS,		0x00c,0x00,0x00,0x00,0x00},
	{"retp",NO_OPS,		0x00d,0x00,0x00,0x00,0x00},
	{"reti",NO_OPS,		0x00e,0x00,0x00,0x00,0x00},
	{"retiw",NO_OPS,	0x00f,0x00,0x00,0x00,0x00},
	{"retw",L,			0x800,0x00,0x00,0xFF,0x00},
	
	{"bank",L,			0x018,0x00,0x00,0xe0,0x05},
	{"page",L,			0x010,0x00,0x00,0xe00,0x09},
	
	{"mode",W,			0x043,0x00,0x00,0x00,0x00},
	{"mode",L,			0x050,0x00,0x00,0x0f,0x00},

	{"option",W,		0x002,0x00,0x00,0x00,0x00},
	
	{"tris",FR_W,		0x000,0x00,0x07,0x00,0x00},
	
	{"iread",NO_OPS,	0x041,0x00,0x00,0x00,0x00},
	{"sleep",NO_OPS,	0x003,0x00,0x00,0x00,0x00},
	
	{"nop",NO_OPS,		0x000,0x00,0x00,0x00,0x00},
	{"",NO_OPS,			0x000,0x00,0x00,0x00,0x00}
};

instruction instructions[MAX_INSTRUCTIONS];
opcode dwOpcode={"dw",NO_OPS,		0x000,0x00,0x00,0x00};

label labels[MAX_LABELS];

unsigned int PC=0;
unsigned int MC=0;
unsigned int lastLabel=0;


opcode *findOpcode(char *mneumonic,int form)
{
	opcode *opcode=opcodes;
	while(opcode->mneumonic[0])
	{
		if((!strcmp(opcode->mneumonic,mneumonic)) && (opcode->form==form))
		{
			return opcode;
		};
		opcode++;
	};
	return NULL;
};

label *findLabel(char *name)
{
	int i;
	for(i=0;i<MAX_LABELS;i++)
	{
		if(!strcmp(name,labels[i].name)) return &labels[i];
	};
	return NULL;
};

void writeHex(FILE *f)
{
	int i=0;
	int c;
	u8 checksum;
	u16 value;
	while(i<MAX_INSTRUCTIONS)
	{
		checksum=0x10+((i & 0xff00) >> 8) + (i & 0xff); 
		fprintf(f,":10%04x00",i*2);
		for(c=0;c<16;c++)
		{
			if(instructions[i].opcode && (instructions[i].opcode!=&dwOpcode))
			{
				value=(instructions[i].opcode->opcode | ((instructions[i].b << 5) & instructions[i].opcode->bMask) | (instructions[i].f & instructions[i].opcode->fMask) | (((instructions[i].l) << (instructions[i].opcode->kShift)) & instructions[i].opcode->kMask) | (instructions[i].reference?(((instructions[i].reference->value) << (instructions[i].opcode->kShift) & instructions[i].opcode->kMask)):0));
				checksum+=((value & 0xff00) >> 8) + (value & 0xff);
				fprintf(f,"%04x",value);
			}
			else
			{
				value=(instructions[i].l & 0x0FFF);
				checksum+=((value & 0xff00) >> 8) + (value & 0xff);
				fprintf(f,"%04x",value);
			};
			i++;
		};
		checksum=-checksum;
		fprintf(f,"%02x\n",checksum);
	};
	fprintf(f,":00000001FF\n");
};

void writeList(FILE *f)
{
	int i;
	u16 value;
		
	for(i=0;i<lastLabel;i++)
	{
		fprintf(f,"%s= $%x = %d\n",labels[i].name,labels[i].value,labels[i].value);
	};
	
	fprintf(f,"\n\n");
	
	i=0;
	while(i<MAX_INSTRUCTIONS)
	{
		if(instructions[i].opcode)
		{
				if(instructions[i].opcode==&dwOpcode)
				{
					value=(instructions[i].l & 0x0FFF);
					if(value) fprintf(f,"%04x:\t%03x\tdw\n",i,value);
				}
				else
				{
					value=(instructions[i].opcode->opcode | ((instructions[i].b << 5) & instructions[i].opcode->bMask) | (instructions[i].f & instructions[i].opcode->fMask) | (((instructions[i].l) << (instructions[i].opcode->kShift)) & instructions[i].opcode->kMask) | (instructions[i].reference?(((instructions[i].reference->value) << (instructions[i].opcode->kShift) & instructions[i].opcode->kMask)):0));
					fprintf(f,"%04x:\t%03x\t%s\t%s\n",i,value,instructions[i].opcode->mneumonic,formStrings[instructions[i].opcode->form]);
				};
		}
		else
		{
			value=(instructions[i].l & 0x0FFF);
			if(value) fprintf(f,"%04x:\t%03x\n",i,value);
		};
		i++;
	};
};

