%{

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "SXAsmCore.h"

//#define YYDEBUG 1

extern int yylineno;

void yyerror(const char *str)
{
	fprintf(stderr,"On line %d error: %s\n",yylineno,str);
};

int yywrap()
{
	return 1;
};

int parse()
{
	return yyparse();
};

label *foundLabel;

%}

%token <longVal> NUMBER 
%token <stringVal> LABELDEF NAME
%token ORG FORG DW DS EQU EOL END

%union
{
	unsigned int longVal;
	char stringVal[256];
}

%type <longVal> expression statement

%left SHL SHR '&' '|'
%left '+' '-'
%left '*' '/'

%start program

%%

program:
			statements END
			{
				YYACCEPT;
			};

statements:		statement  { };
				| statements statement { };

expression:
				NUMBER
				{
					$$=$1; 
				};
expression:				
				NAME
				{
					struct label *foundLabel=findLabel($1);
					if(!foundLabel) { fprintf(stderr,"On line %d error: Undefined idenfifier: %s. (0)",yylineno,$1); exit(1); };
					$$=foundLabel->value;
				};
				
expression:
				expression '+' expression
				{
					$$=$1+$3;
				};
				
expression:
				expression '-' expression
				{
					$$=$1-$3;
				};
				
expression:
				expression '*' expression
				{
					$$=$1*$3;
				};
				
expression:
				expression '/' expression
				{
					$$=$1/$3;
				};
				
expression:
				expression SHL expression
				{
					$$=$1<<$3;
				};
				
expression:
				expression SHR expression
				{
					$$=$1>>$3;
				};
				
expression:
				expression '&' expression
				{
					$$=$1&$3;
				};
				
expression:
				expression '|' expression
				{
					$$=$1|$3;
				};
				
expression:
				'(' expression ')'
				{
					$$=$2;
				};

statement:
				EOL 
				{
					$$=0;
				};

statement:
				ORG expression EOL 
				{
					$$=0;
					PC=$2;
				};
				
statement:
				FORG expression EOL 
				{
					$$=0;
					MC=$2;
				};
				
statement:
				DW expression EOL 
				{
					$$=PC;
					instructions[$$].opcode=&dwOpcode;
					instructions[$$].b=0;
					instructions[$$].f=0;
					instructions[$$].l=$2;
					instructions[$$].reference=NULL;
					PC++;
				};
				
statement:
				LABELDEF DS EOL 
				{
					$$=0;
					foundLabel=findLabel($1);
					if(foundLabel)
					{
						foundLabel->value=MC;
					}
					else
					{
						strcpy(labels[lastLabel].name,$1);
						labels[lastLabel].value=MC;
						lastLabel++;
					};
					MC++;
				};
				
statement:
				NAME EOL 
				{
					$$=PC;
					if(!(instructions[$$].opcode=findOpcode($1,NO_OPS))) { fprintf(stderr,"On line %d error: Unknown opcode or form: %s. (1)",yylineno,$1); exit(1); };
					instructions[$$].b=0;
					instructions[$$].f=0;
					instructions[$$].l=0;
					instructions[$$].reference=NULL;
					PC++;
				};

statement:
				NAME 'w' EOL 
				{
					$$=PC;
					if(!(instructions[$$].opcode=findOpcode($1,W))) { fprintf(stderr,"On line %d error: Unknown opcode or form: %s. (2)",yylineno,$1); exit(1); };
					instructions[$$].b=0;
					instructions[$$].f=0;
					instructions[$$].l=0;
					instructions[$$].reference=NULL;
					PC++;
				};
				
statement:
				NAME expression EOL 
				{
					$$=PC;
					if(!(instructions[$$].opcode=findOpcode($1,FR))) { fprintf(stderr,"On line %d error: Unknown opcode or form: %s. (3)",yylineno,$1); exit(1); };
					instructions[$$].b=0;
					instructions[$$].f=$2;
					instructions[$$].l=0;
					instructions[$$].reference=NULL;
					PC++;
				};
								
statement:
				NAME '#' expression EOL 
				{
					$$=PC;
					if(!(instructions[$$].opcode=findOpcode($1,L))) { fprintf(stderr,"On line %d error: Unknown opcode or form: %s. (4)",yylineno,$1); exit(1); };
					instructions[$$].b=0;
					instructions[$$].f=0;
					instructions[$$].l=$3;
					instructions[$$].reference=NULL;
					PC++;
				};
								
statement:
				NAME expression ',' 'w' EOL 
				{
					$$=PC;
					if(!(instructions[$$].opcode=findOpcode($1,FR_W))) { fprintf(stderr,"On line %d error: Unknown opcode or form: %s. (5)",yylineno,$1); exit(1); };
					instructions[$$].b=0;
					instructions[$$].f=$2;
					instructions[$$].l=0;
					instructions[$$].reference=NULL;
					PC++;
				};
												
statement:
				NAME  'w' ',' expression EOL 
				{
					$$=PC;
					if(!(instructions[$$].opcode=findOpcode($1,W_FR))) { fprintf(stderr,"On line %d error: Unknown opcode or form: %s.  (6)",yylineno,$1); exit(1); };
					instructions[$$].b=0;
					instructions[$$].f=$4;
					instructions[$$].l=0;
					instructions[$$].reference=NULL;
					PC++;
				};
				
statement:
				NAME  'w' ',' '#' expression EOL 
				{
					$$=PC;
					if(!(instructions[$$].opcode=findOpcode($1,W_L))) { fprintf(stderr,"On line %d error: Unknown opcode or form: %s. (7)",yylineno,$1); exit(1); };
					instructions[$$].b=0;
					instructions[$$].f=0;
					instructions[$$].l=$5;
					instructions[$$].reference=NULL;
					PC++;
				};
				
statement:
				NAME expression '.' expression EOL 
				{
					$$=PC;
					if(!(instructions[$$].opcode=findOpcode($1,FR_B))) { fprintf(stderr,"On line %d error: Unknown opcode or form: %s. (7)",yylineno,$1); exit(1); };
					instructions[$$].b=$4;
					instructions[$$].f=$2;
					instructions[$$].l=0;
					instructions[$$].reference=NULL;
					PC++;
				};
				
statement:
				NAME '@' NAME EOL 
				{
					$$=PC;
					if(!(instructions[$$].opcode=findOpcode($1,L))) { fprintf(stderr,"On line %d error: Unknown opcode or form: %s. (8)",yylineno,$1); exit(1); };
					instructions[$$].b=0;
					instructions[$$].f=0;
					instructions[$$].l=0;
					if(!(instructions[$$].reference=findLabel($3))) 
					{
						strcpy(labels[lastLabel].name,$3);
						labels[lastLabel].value=0;
						lastLabel++;
					};
					PC++;
				};
												
statement:
				LABELDEF statement 
				{
					$$=$2;
					foundLabel=findLabel($1);
					if(foundLabel)
					{
						foundLabel->value=$$;
					}
					else
					{
						strcpy(labels[lastLabel].name,$1);
						labels[lastLabel].value=$$;
						lastLabel++;
					};
				};
				
statement:
				LABELDEF EQU expression EOL 
				{
					$$=0;
					foundLabel=findLabel($1);
					if(foundLabel)
					{
						foundLabel->value=$3;
					}
					else
					{
						strcpy(labels[lastLabel].name,$1);
						labels[lastLabel].value=$3;
						lastLabel++;
					};
				};
statement:
				LABELDEF EQU '@' EOL 
				{
					$$=0;
					foundLabel=findLabel($1);
					if(foundLabel)
					{
						foundLabel->value=PC;
					}
					else
					{
						strcpy(labels[lastLabel].name,$1);
						labels[lastLabel].value=PC;
						lastLabel++;
					};
				};
				
				
				
%%