/*
 *  SXCodeGen.c
 *  mcc
 *
 *  Created by Britt Durbrow on 7/17/06.
 *  Copyright 2006 Rattlesnake Hill Softworks. All rights reserved.
 *
 */

#include "SXCodeGen.h"
#include "mccCore.h"
#include <string.h>

#define YES 0xFF
#define NO 0x00

int llabel;

char vdxVariables[16][32];
int vdxVariableCount;

int compileExpression(FILE *f, operation *op, int gscratch, int lscratch, int lvar,int isVdx,int generateResult)
{
	variable *var;
	operation *parOp;
	int i;
	
	switch(op->op)
	{
		case varRefOp:
			var=parsedProgram->globals;
			while(var)
			{
				if(strcmp(op->name,var->name)==0)
				{
					// Compile a global refrence.
					fprintf(f,"\t\tmov w,FSR;\n");
					fprintf(f,"\t\tmov $0f,w;\n");
					fprintf(f,"\t\tbank #1;\n");
					fprintf(f,"\t\tmov w,%s;\n",op->name);
					fprintf(f,"\t\tmov $0e,w;\n");
					fprintf(f,"\t\tmov w,$0f;\n");
					fprintf(f,"\t\tmov FSR,w;\n");
					fprintf(f,"\t\tmov w,$0e;\n");
					return 0;
				};
				var=var->prev;
			};
			
			if(!isVdx)
			{
				i=0;
				while(i<vdxVariableCount)
				{
					if(strcmp(op->name,vdxVariables[i])==0)
					{
						// Compile a vdx refrence.
						fprintf(f,"\t\tmov w,FSR;\n");
						fprintf(f,"\t\tmov $0f,w;\n");
						fprintf(f,"\t\tbank #0;\n");
						fprintf(f,"\t\tmov w,vdx_%s;\n",op->name);
						fprintf(f,"\t\tmov $0e,w;\n");
						fprintf(f,"\t\tmov w,$0f;\n");
						fprintf(f,"\t\tmov FSR,w;\n");
						fprintf(f,"\t\tmov w,$0e;\n");
						return 0;
					};
					i++;
				};
			};
			
			// it wasn't a global or an external vdx variable, so... compile a local refrence.
			if(isVdx)
			{
				fprintf(f,"\t\tmov w,vdx_%s;\n",op->name);
			}
			else
			{
				fprintf(f,"\t\tmov w,__%s;\n",op->name);
			};
		return 0;


		case varAddrOp:
			var=parsedProgram->globals;
			while(var)
			{
				if(strcmp(op->name,var->name)==0)
				{
					// Compile a global refrence.
					fprintf(f,"\t\tmov w,#%s;\n",op->name);			
					return 0;
				};
				var=var->prev;
			};
			
			if(!isVdx)
			{
				i=0;
				while(i<vdxVariableCount)
				{
					if(strcmp(op->name,vdxVariables[i])==0)
					{
						// Compile a vdx refrence.
						fprintf(f,"\t\tmov w,#vdx_%s;\n",op->name);
						return 0;
					};
					i++;
				};
			};
			
			// it wasn't a global or an external vdx variable, so... compile a local refrence.
			if(isVdx)
			{
				fprintf(f,"\t\tmov w,#vdx_%s;\n",op->name);
			}
			else
			{
				fprintf(f,"\t\tmov w,#__%s;\n",op->name);
			};
		return 0;


		case constOp:
			fprintf(f,"\t\tmov w,#$%02x;\n",op->value);
		return 0;


		case addOp:
			compileExpression(f,op->right,gscratch,lscratch,lvar,isVdx,generateResult);
			fprintf(f,"\t\tmov $%02x,w;\n",lscratch);
			compileExpression(f,op->left,gscratch,lscratch-1,lvar,isVdx,generateResult);
			fprintf(f,"\t\tadd w,$%02x;\n",lscratch);
		return 0;


		case subOp:
			compileExpression(f,op->left,gscratch,lscratch,lvar,isVdx,generateResult);
			fprintf(f,"\t\tmov $%02x,w;\n",lscratch);
			compileExpression(f,op->right,gscratch,lscratch-1,lvar,isVdx,generateResult);
			fprintf(f,"\t\tsub w,$%02x;\n",lscratch);
		return 0;


		case mulOp:
		return 0;


		case divOp:
		return 0;


		case gtOp:
			compileExpression(f,op->right,gscratch,lscratch,lvar,isVdx,generateResult);
			fprintf(f,"\t\tmov $%02x,w;\n",lscratch);
			compileExpression(f,op->left,gscratch,lscratch-1,lvar,isVdx,generateResult);
			fprintf(f,"\t\tsub w,$%02x;\n",lscratch);
			fprintf(f,"\t\tmov w,#$ff;\n");
			fprintf(f,"\t\tsc;\n");
			fprintf(f,"\t\tmov w,#$00;\n");
		return 0;


		case ltOp:
			compileExpression(f,op->left,gscratch,lscratch,lvar,isVdx,generateResult);
			fprintf(f,"\t\tmov $%02x,w;\n",lscratch);
			compileExpression(f,op->right,gscratch,lscratch-1,lvar,isVdx,generateResult);
			fprintf(f,"\t\tsub w,$%02x;\n",lscratch);
			fprintf(f,"\t\tmov w,#$ff;\n");
			fprintf(f,"\t\tsc;\n");
			fprintf(f,"\t\tmov w,#$00;\n");
		return 0;


		case equOp:
			compileExpression(f,op->right,gscratch,lscratch,lvar,isVdx,generateResult);
			fprintf(f,"\t\tmov $%02x,w;\n",lscratch);
			compileExpression(f,op->left,gscratch,lscratch-1,lvar,isVdx,generateResult);
			fprintf(f,"\t\tsub w,$%02x;\n",lscratch);
			fprintf(f,"\t\tmov w,#$ff;\n");
			fprintf(f,"\t\tsz;\n");
			fprintf(f,"\t\tmov w,#$00;\n");
		return 0;


		case landOp:
			compileExpression(f,op->left,gscratch,lscratch,lvar,isVdx,generateResult);
			fprintf(f,"\t\tmov $%02x,w;\n",lscratch);
			compileExpression(f,op->right,gscratch,lscratch-1,lvar,isVdx,generateResult);
			fprintf(f,"\t\ttest $%02x;\n",lscratch);
			fprintf(f,"\t\tsz;\n");
			fprintf(f,"\t\tmov w,#$00;\n");
		return 0;


		case lorOp:
			compileExpression(f,op->left,gscratch,lscratch,lvar,isVdx,generateResult);
			fprintf(f,"\t\tmov $%02x,w;\n",lscratch);
			compileExpression(f,op->right,gscratch,lscratch-1,lvar,isVdx,generateResult);
			fprintf(f,"\t\ttest $%02x;\n",lscratch);
			fprintf(f,"\t\tsz;\n");
			fprintf(f,"\t\tmov w,#ff;\n");
		return 0;


		case lnotOp:
			compileExpression(f,op->right,gscratch,lscratch,lvar,isVdx,generateResult);
			fprintf(f,"\t\ttest w;\n",lscratch);
			fprintf(f,"\t\tmov w,#$ff;\n");
			fprintf(f,"\t\tsz;\n");
			fprintf(f,"\t\tmov w,#$00;\n");
		return 0;


		case assignOp:
			compileExpression(f,op->right,gscratch,lscratch,lvar,isVdx,generateResult);
			
			var=parsedProgram->globals;
			while(var)
			{
				if(strcmp(op->name,var->name)==0)
				{
					// Compile a global refrence.
					fprintf(f,"\t\tmov $0e,w;\n");
					fprintf(f,"\t\tmov w,FSR;\n");
					fprintf(f,"\t\tmov $0f,w;\n");
					fprintf(f,"\t\tbank #1;\n");
					fprintf(f,"\t\tmov w,$0e;\n");
					fprintf(f,"\t\tmov %s,w;\n",op->name);
					fprintf(f,"\t\tmov w,$0f;\n");
					fprintf(f,"\t\tmov FSR,w;\n");
					if(generateResult) fprintf(f,"\t\tmov w,$0e;\n");
					return 0;
				};
				var=var->prev;
			};
			
			if(!isVdx)
			{
				i=0;
				while(i<vdxVariableCount)
				{
					if(strcmp(op->name,vdxVariables[i])==0)
					{
						// Compile a vdx refrence.						
						fprintf(f,"\t\tmov $0e,w;\n");
						fprintf(f,"\t\tmov w,FSR;\n");
						fprintf(f,"\t\tmov $0f,w;\n");
						fprintf(f,"\t\tbank #0;\n");
						fprintf(f,"\t\tmov w,$0e;\n");
						fprintf(f,"\t\tmov vdx_%s,w;\n",op->name);
						fprintf(f,"\t\tmov w,$0f;\n");
						fprintf(f,"\t\tmov FSR,w;\n");
						if(generateResult) fprintf(f,"\t\tmov w,$0e;\n");						
						return 0;
					};
					i++;
				};
			};
			
			// it wasn't a global or an external vdx variable, so... compile a local refrence.
			if(isVdx)
			{
				fprintf(f,"\t\tmov vdx_%s,w;\n",op->name);
			}
			else
			{
				fprintf(f,"\t\tmov __%s,w;\n",op->name);
			};
		return 0;


		case bandOp:
			compileExpression(f,op->right,gscratch,lscratch,lvar,isVdx,generateResult);
			fprintf(f,"\t\tmov $%02x,w;\n",lscratch);
			compileExpression(f,op->left,gscratch,lscratch-1,lvar,isVdx,generateResult);
			fprintf(f,"\t\tand w,$%02x;\n",lscratch);
		return 0;


		case borOp:
			compileExpression(f,op->right,gscratch,lscratch,lvar,isVdx,generateResult);
			fprintf(f,"\t\tmov $%02x,w;\n",lscratch);
			compileExpression(f,op->left,gscratch,lscratch-1,lvar,isVdx,generateResult);
			fprintf(f,"\t\tor w,$%02x;\n",lscratch);
		return 0;


		case bnotOp:
			compileExpression(f,op->right,gscratch,lscratch,lvar,isVdx,generateResult);
			fprintf(f,"\t\txor w,#$ff;\n");
		return 0;


		case bxorOp:
			compileExpression(f,op->right,gscratch,lscratch,lvar,isVdx,generateResult);
			fprintf(f,"\t\tmov $%02x,w;\n",lscratch);
			compileExpression(f,op->left,gscratch,lscratch-1,lvar,isVdx,generateResult);
			fprintf(f,"\t\txor w,$%02x;\n",lscratch);
		return 0;


		case shlOp:
			compileExpression(f,op->right,gscratch,lscratch,lvar,isVdx,generateResult);
			fprintf(f,"\t\tmov $%02x,w;\n",lscratch);
			fprintf(f,"\t\tclc;\n");
			fprintf(f,"\t\trl w,$%02x;\n",lscratch);
		return 0;


		case shrOp:
			compileExpression(f,op->right,gscratch,lscratch,lvar,isVdx,generateResult);
			fprintf(f,"\t\tmov $%02x,w;\n",lscratch);
			fprintf(f,"\t\tclc;\n");
			fprintf(f,"\t\trr w,$%02x;\n",lscratch);
		return 0;
		
		case incOp:
			var=parsedProgram->globals;
			while(var)
			{
				if(strcmp(op->name,var->name)==0)
				{
					// Compile a global refrence.
					fprintf(f,"\t\tmov w,FSR;\n");
					fprintf(f,"\t\tmov $0f,w;\n");
					fprintf(f,"\t\tbank #1;\n");
					fprintf(f,"\t\tinc %s;\n",op->name);
					if(generateResult)
					{
						fprintf(f,"\t\tmov w,%s;\n");
						fprintf(f,"\t\tmov $0e,w;\n");
					};
					fprintf(f,"\t\tmov w,$0f;\n");
					fprintf(f,"\t\tmov FSR,w;\n");
					if(generateResult) fprintf(f,"\t\tmov w,$0e;\n");
					return 0;
				};
				var=var->prev;
			};
			
			if(!isVdx)
			{
				i=0;
				while(i<vdxVariableCount)
				{
					if(strcmp(op->name,vdxVariables[i])==0)
					{
						// Compile a vdx refrence.						
						fprintf(f,"\t\tmov w,FSR;\n");
						fprintf(f,"\t\tmov $0f,w;\n");
						fprintf(f,"\t\tbank #0;\n");
						fprintf(f,"\t\tinc vdx_%s;\n",op->name);
						if(generateResult) 
						{
							fprintf(f,"\t\tmov w,vdx_%s;\n");
							fprintf(f,"\t\tmov $0e,w;\n");
						};
						fprintf(f,"\t\tmov w,$0f;\n");
						fprintf(f,"\t\tmov FSR,w;\n");
						if(generateResult) fprintf(f,"\t\tmov w,$0e;\n");
						return 0;
					};
					i++;
				};
			};
			
			// it wasn't a global or an external vdx variable, so... compile a local refrence.
			if(isVdx)
			{
				fprintf(f,"\t\tinc vdx_%s;\n",op->name);
				if(generateResult) fprintf(f,"\t\tmov w,vdx_%s;\n",op->name);
			}
			else
			{
				fprintf(f,"\t\tinc __%s;\n",op->name);
				if(generateResult) fprintf(f,"\t\tmov w,__%s;\n",op->name);
			};
		return 0;
		
		case decOp:
			var=parsedProgram->globals;
			while(var)
			{
				if(strcmp(op->name,var->name)==0)
				{
					// Compile a global refrence.
					fprintf(f,"\t\tmov w,FSR;\n");
					fprintf(f,"\t\tmov $0f,w;\n");
					fprintf(f,"\t\tbank #1;\n");
					fprintf(f,"\t\tdec %s;\n",op->name);
					if(generateResult)
					{
						fprintf(f,"\t\tmov w,%s;\n");
						fprintf(f,"\t\tmov $0e,w;\n");
					};
					fprintf(f,"\t\tmov w,$0f;\n");
					fprintf(f,"\t\tmov FSR,w;\n");
					if(generateResult) fprintf(f,"\t\tmov w,$0e;\n");
					return 0;
				};
				var=var->prev;
			};
			
			if(!isVdx)
			{
				i=0;
				while(i<vdxVariableCount)
				{
					if(strcmp(op->name,vdxVariables[i])==0)
					{
						// Compile a vdx refrence.						
						fprintf(f,"\t\tmov w,FSR;\n");
						fprintf(f,"\t\tmov $0f,w;\n");
						fprintf(f,"\t\tbank #0;\n");
						fprintf(f,"\t\tdec vdx_%s;\n",op->name);
						if(generateResult)
						{
							fprintf(f,"\t\tmov w,vdx_%s;\n");
							fprintf(f,"\t\tmov $0e,w;\n");
						};
						fprintf(f,"\t\tmov w,$0f;\n");
						fprintf(f,"\t\tmov FSR,w;\n");
						if(generateResult) fprintf(f,"\t\tmov w,$0e;\n");
						return 0;
					};
					i++;
				};
			};
			
			// it wasn't a global or an external vdx variable, so... compile a local refrence.
			if(isVdx)
			{
				fprintf(f,"\t\tdec vdx_%s;\n",op->name);
				if(generateResult) fprintf(f,"\t\tmov w,vdx_%s;\n",op->name);
			}
			else
			{
				fprintf(f,"\t\tdec __%s;\n",op->name);
				if(generateResult) fprintf(f,"\t\tmov w,__%s;\n",op->name);
			};
		return 0;
		
		case callOp:
			parOp=op->right;
			if(parOp)
			{
				i=8;
				while(parOp->prev) parOp=parOp->prev;
				while(parOp)
				{
					compileExpression(f,parOp,gscratch,lscratch,lvar,isVdx,generateResult);
					fprintf(f,"\t\tmov $%02x,w;\n",i);
					i++;
					parOp=parOp->next;
				};
			};
			
			fprintf(f,"\t\tpage #0;\n");
			fprintf(f,"\t\tcall #%s;\n",op->name);
			if(generateResult) fprintf(f,"\t\tmov w,$08;\n");
		return 0;
	};

	return 0;
};

int compileStatement(FILE *f, statement *st, int *lvar, int isVdx)
{
	statement *subSt;

	fprintf(f,"\n");
	switch(st->construct)
	{
		case asmStatement:
			fprintf(f,"\n%s\n",st->text);
		return 0;
	
		case exprStatement:
			compileExpression(f,st->op,0x08,0x1f,*lvar,isVdx,NO);
				fprintf(f,"\n");
		return 0;


		case ifStatement:
			if(st->lower)
			{
				compileExpression(f,st->op,0x08,0x1f,*lvar,isVdx,YES);
				fprintf(f,"\t\ttest w;\n");
				fprintf(f,"\t\tpage @___else%d;\n",llabel);
				fprintf(f,"\t\tsz;\n");
				fprintf(f,"\t\tjmp @___else%d;\n",llabel);
				
				if(subSt=st->upper)
				{
					while(subSt->prev) subSt=subSt->prev;
					
					while(subSt)
					{
						compileStatement(f,subSt,lvar,isVdx);
						subSt=subSt->next;
					};
					fprintf(f,"\n");
				};
				fprintf(f,"\t\tpage @___endif%d;\n",llabel);
				fprintf(f,"\t\tjmp @___endif%d;\n",llabel);
				fprintf(f,"___else%d:\tequ @;\n",llabel);
				
				if(subSt=st->lower)
				{
					while(subSt->prev) subSt=subSt->prev;
					
					while(subSt)
					{
						compileStatement(f,subSt,lvar,isVdx);
						subSt=subSt->next;
					};
					fprintf(f,"\n");
				};
				
				fprintf(f,"___endif%d:\tequ @;\n",llabel);
			}
			else
			{
				compileExpression(f,st->op,0x08,0x1f,*lvar,isVdx,YES);
				fprintf(f,"\t\ttest w;\n");
				fprintf(f,"\t\tpage @___endif%d;\n",llabel);
				fprintf(f,"\t\tsz;\n");
				fprintf(f,"\t\tjmp @___endif%d;\n",llabel);
				
				if(subSt=st->upper)
				{
					while(subSt->prev) subSt=subSt->prev;
					
					while(subSt)
					{
						compileStatement(f,subSt,lvar,isVdx);
						subSt=subSt->next;
					};
					fprintf(f,"\n");
				};
				
				fprintf(f,"___endif%d:\tequ @;\n",llabel);
			};
			llabel++;
		return 0;


		case whileStatement:
			fprintf(f,"___while%d:\tequ @;\n",llabel);
			compileExpression(f,st->op,0x08,0x1f,*lvar,isVdx,YES);
			fprintf(f,"\t\ttest w;\n");
			fprintf(f,"\t\tpage @___endwhile%d;\n",llabel);
			fprintf(f,"\t\tsnz;\n");
			fprintf(f,"\t\tjmp @___endwhile%d;\n",llabel);
			
			if(subSt=st->upper)
			{
				while(subSt->prev) subSt=subSt->prev;
				
				while(subSt)
				{
					compileStatement(f,subSt,lvar,isVdx);
					subSt=subSt->next;
				};
				fprintf(f,"\n");
			};
			fprintf(f,"\t\tpage @___while%d;\n",llabel);
			fprintf(f,"\t\tjmp @___while%d;\n",llabel);
			fprintf(f,"___endwhile%d:\tequ @;\n",llabel);
			llabel++;
		return 0;


		case repeatStatement:
			fprintf(f,"___repeat%d:\tequ @;\n",llabel);
			
			if(subSt=st->upper)
			{
				while(subSt->prev) subSt=subSt->prev;
				
				while(subSt)
				{
					compileStatement(f,subSt,lvar,isVdx);
					subSt=subSt->next;
				};
				fprintf(f,"\n");
			};
			
			compileExpression(f,st->op,0x08,0x1f,*lvar,isVdx,YES);
			fprintf(f,"\t\ttest w;\n");
			fprintf(f,"\t\tpage @___repeat%d;\n",llabel);
			fprintf(f,"\t\tsz;\n");
			fprintf(f,"\t\tjmp @___repeat%d;\n",llabel);
			llabel++;
		return 0;
		
		case returnStatement:
			compileExpression(f,st->op,0x08,0x1f,*lvar,isVdx,YES);
			fprintf(f,"\t\tmov $08,w;\n");
			fprintf(f,"\t\tret;\n\n");
		return 0;

		case declarationStatement:
			if(isVdx)
			{
				strcpy(vdxVariables[vdxVariableCount],st->var->name);
				vdxVariableCount++;
				fprintf(f,"vdx_%s:\tds\n",st->var->name);
			}
			else
			{
				fprintf(f,"__%s:\tds\n",st->var->name);
			};
			(*lvar)++;
		return 0;
	};
	
	return 0;
};

int compileFunction(FILE *f,function *func)
{
	statement *st;
	variable *par;
	int i;
	int lvar=0x10;
	
	// Function prolog
	fprintf(f,"\n\n_%s:\tmov w,#$20;\n",func->name);
	fprintf(f,"\t\tadd FSR,w;\n");
	
	// move parameters to locals
	i=8;
	fprintf(f,"\t\tforg $10;\n");
	par=func->parameters;
	if(par)
	{
		while(par->prev) par=par->prev;
		while(par)
		{
			fprintf(f,"__%s:\tds;\n",par->name);
			fprintf(f,"\t\tmov w,$%02x;\n",i);
			fprintf(f,"\t\tmov __%s,w;\n",par->name);
			par=par->next;
			i++;
			lvar++;
		};
	};
		
	// rewind st; as the parser returns the end of the list
	st=func->statements;
	if(st)
		while(st->prev) st=st->prev;
	
	// compile the statements
	while(st)
	{
		compileStatement(f,st,&lvar,NO);
		st=st->next;
	};
	
	fprintf(f,"\t\tclr $08;\n");
	fprintf(f,"\t\tret;\n\n");

	
	return 0;
};

int compileModule(FILE *f,function *func,int *lvar)
{
	statement *st;
	int isVdx=func->kind==vdxKind;
	
	// rewind st; as the parser returns the end of the list
	st=func->statements;
	if(st)
		while(st->prev) st=st->prev;
	
	// compile the statements
	while(st)
	{
		compileStatement(f,st,lvar,isVdx);
		st=st->next;
	};
	
	return 0;
};


int compileSX(FILE *f)
{
	function *func;
	variable *var;
	int lvar;

	llabel=0;
	
	fprintf(f,"FSR:\tequ\t$04;\n");
	fprintf(f,"__PORTA:\tequ\t$05;\n");
	fprintf(f,"__PORTB:\tequ\t$06;\n");
	fprintf(f,"__PORTC:\tequ\t$07;\n");

	fprintf(f,"\n\n");
	
	fprintf(f,"\t\torg 0;\n");
	fprintf(f,"\t\tpage @vdx;\n");
	fprintf(f,"\t\tjmp @vdx;\n\n");
	
	fprintf(f,"reset:\tpage @init;\n");
	fprintf(f,"\t\tjmp @init;\n\n");
	
	fprintf(f,"\t\torg 2047;\n");
	fprintf(f,"\t\tjmp @reset;\n");
	
	fprintf(f,"\n\n; Globals\n\n");
	fprintf(f,"\t\tforg $10;\n");
	var=parsedProgram->globals;
	while(var)
	{
		fprintf(f,"%s:\tds;\n",var->name);		
		var=var->prev;
	};
	
	fprintf(f,"\n; Function Jump Table\n\n");
	fprintf(f,"\t\torg $10;\n");
	func=parsedProgram->functions;
	if(func)
	{
		while(func->prev) func=func->prev;
		while(func)
		{
			if(func->kind==functionKind)
			{
				fprintf(f,"%s:\tpage @_%s;\n",func->name,func->name);
				fprintf(f,"\t\tjmp @_%s;\n",func->name,func->name);
			};
			func=func->next;
		};
	};
	
	vdxVariableCount=0;
	
	fprintf(f,"\n; VDx\n\n");
	fprintf(f,"vdx:\tequ @;\n");
	fprintf(f,"\t\tbank #$00;\n");
	func=parsedProgram->functions;
	if(func)
	{
		while(func->prev) func=func->prev;
		while(func)
		{
			if(func->kind==vdxKind) compileModule(f,func,&lvar);
			func=func->next;
		};
	};
	fprintf(f,"\t\treti;\n\n");

	fprintf(f,"\n; Init\n\n");
	fprintf(f,"init:\tequ @;\n");
	fprintf(f,"\t\tbank #$02;\n");
	func=parsedProgram->functions;
	if(func)
	{
		while(func->prev) func=func->prev;
		while(func)
		{
			if(func->kind==initKind) compileModule(f,func,&lvar);
			func=func->next;
		};
	};
	fprintf(f,"\t\tpage @_main;\n");
	fprintf(f,"\t\tjmp @_main;\n");
	
	fprintf(f,"\n; Functions\n\n");
	func=parsedProgram->functions;
	if(func)
	{
		while(func->prev) func=func->prev;
		while(func)
		{
			if(func->kind==functionKind) compileFunction(f,func);
			func=func->next;
		};
	};
	
	fprintf(f,"\n\nend\n\n");
	return 0;
};