%{

#include <stdio.h>
#include <string.h>
#include "mccCore.h"

//#define YYDEBUG 1

variable *var;
function *func;

void yyerror(const char *str)
{
	fprintf(stderr,"error: %s\n",str);
	exit(1);
};

int yywrap()
{
	return 1;
};

int parse()
{
	return yyparse();
};

%}

%union
{
	unsigned int longVal;
	char stringVal[STRINGBUFFERSIZE];
	variable *varVal;
	operation *opVal;
	statement *stVal;
	function *funcVal;
};

%token <stringVal> NAME ASMBLOCK
%token <longVal> LONG
%token IF WHILE REPEAT RETURN INIT VDX ASSIGN SHL SHR BAND BOR INCOP DECOP BYTE WORD QUADWORD 

%type <opVal> expressions expression
%type <stVal> statements statement
%type <varVal> globals global parameters parameter
%type <funcVal> functions function

%left '(' ')'
%left ASSIGN
%left  '&' '|' '!'
%left '=' '<' '>'
%left SHL SHR BAND BOR '~' '^'
%left '+' '-'
%left '*' '/'
%left INCOP DECOP

%start program

%%

program:
				NAME '{' globals functions '}' ';'
				{
					if(($3!=NULL) && (parsedProgram->globals!=NULL))
					{
						var=$3;
						while(var->prev) var=var->prev;
						
						var->prev=parsedProgram->globals;
						parsedProgram->globals->next=var;
					};
					if($3!=NULL) parsedProgram->globals=$3;

					
					
					if(($4!=NULL) && (parsedProgram->functions!=NULL))
					{
						func=$4;
						while(func->prev) func=func->prev;
						
						func->prev=parsedProgram->functions;
						parsedProgram->functions->next=func;
					};
					if($4!=NULL) parsedProgram->functions=$4;

					
					YYACCEPT;
				};
				
globals:
	/* empty */
	{
		$$=NULL;
	};
		
	| globals global ';'	 
	{
		$$=$2;
		$$->next=NULL;
		$$->prev=$1;
		if($1) $1->next=$$;
	};
	
functions:
	/* empty */
	{
		$$=NULL;
	};
		
	| functions function ';'	 
	{
		$$=$2;
		$$->next=NULL;
		$$->prev=$1;
		if($1) $1->next=$$;
	};
	
parameters:
	/* empty */
	{
		$$=NULL;
	};
	| parameter
	{
		$$=$1;
		$$->next=NULL;
		$$->prev=NULL;
	};
	| parameters ',' parameter
	{
		$$=$3;
		$$->next=NULL;
		$$->prev=$1;
		if($1) $1->next=$$;
	};
	
expressions:
	/* empty */
	{
		$$=NULL;
	};
	| expression
	{
		$$=$1;
		$$->next=NULL;
		$$->prev=NULL;
	};
	| expressions ',' expression
	{
		$$=$3;
		$$->next=NULL;
		$$->prev=$1;
		if($1) $1->next=$$;
	};
	
statements:
	/* empty */
	{
		$$=NULL;
	};
		
	| statements statement ';'	 
	{
		$$=$2;
		$$->next=NULL;
		$$->prev=$1;
		if($1) $1->next=$$;
	};
	
	
global:
				BYTE NAME
				{
					$$=malloc(sizeof(variable));
					$$->name=malloc(NAMESIZE);
					strcpy($$->name,$2);
					$$->type=byte;
					$$->size=1;
					$$->next=NULL;
					$$->prev=NULL;
				};


function:
				NAME '(' parameters ')' '{' statements '}'
				{
					$$=malloc(sizeof(function));
					$$->name=malloc(NAMESIZE);
					strcpy($$->name,$1);
					$$->parameters=$3;
					$$->statements=$6;
					$$->next=NULL;
					$$->prev=NULL;
					$$->kind=functionKind;
				};
				
function:
				INIT NAME '{' statements '}'
				{
					$$=malloc(sizeof(function));
					$$->name=malloc(NAMESIZE);
					strcpy($$->name,$2);
					$$->parameters=NULL;
					$$->statements=$4;
					$$->next=NULL;
					$$->prev=NULL;
					$$->kind=initKind;
				};
				
function:
				VDX NAME '{' statements '}'
				{
					$$=malloc(sizeof(function));
					$$->name=malloc(NAMESIZE);
					strcpy($$->name,$2);
					$$->parameters=NULL;
					$$->statements=$4;
					$$->next=NULL;
					$$->prev=NULL;
					$$->kind=vdxKind;
				};

parameter:
				BYTE NAME
				{
					$$=malloc(sizeof(variable));
					$$->name=malloc(NAMESIZE);
					strcpy($$->name,$2);
					$$->type=byte;
					$$->size=1;
					$$->next=NULL;
					$$->prev=NULL;
				};
				
expression:
				NAME
				{
					$$=malloc(sizeof(operation));
					$$->left=NULL;
					$$->right=NULL;
					$$->next=NULL;
					$$->prev=NULL;
					$$->op=varRefOp;
					$$->value=0;
					$$->name=malloc(NAMESIZE);
					strcpy($$->name,$1);
				};
				
expression:
				'@' NAME
				{
					$$=malloc(sizeof(operation));
					$$->left=NULL;
					$$->right=NULL;
					$$->next=NULL;
					$$->prev=NULL;
					$$->op=varAddrOp;
					$$->value=0;
					$$->name=malloc(NAMESIZE);
					strcpy($$->name,$2);
				};
				
expression:
				LONG
				{
					$$=malloc(sizeof(operation));
					$$->left=NULL;
					$$->right=NULL;
					$$->next=NULL;
					$$->prev=NULL;
					$$->op=constOp;
					$$->value=$1;
					$$->name=NULL;
				};

expression:
				NAME '(' expressions ')'
				{
					$$=malloc(sizeof(operation));
					$$->left=NULL;
					$$->right=$3;
					$$->next=NULL;
					$$->prev=NULL;
					$$->op=callOp;
					$$->value=0;
					$$->name=malloc(NAMESIZE);
					strcpy($$->name,$1);
				};

expression:
				'(' expression ')'
				{
					$$=$2;
				};

expression:
				NAME ASSIGN expression
				{
					$$=malloc(sizeof(operation));
					$$->left=NULL;
					$$->right=$3;
					$$->next=NULL;
					$$->prev=NULL;
					$$->op=assignOp;
					$$->value=0;
					$$->name=malloc(NAMESIZE);
					strcpy($$->name,$1);
				};

expression:
				expression '&' expression
				{
					$$=malloc(sizeof(operation));
					$$->left=$1;
					$$->right=$3;
					$$->next=NULL;
					$$->prev=NULL;
					$$->op=landOp;
					$$->value=0;
					$$->name=NULL;
				};
				
expression:
				expression '|' expression
				{
					$$=malloc(sizeof(operation));
					$$->left=$1;
					$$->right=$3;
					$$->next=NULL;
					$$->prev=NULL;
					$$->op=lorOp;
					$$->value=0;
					$$->name=NULL;
				};

expression:
				'!' expression
				{
					$$=malloc(sizeof(operation));
					$$->left=NULL;
					$$->right=$2;
					$$->next=NULL;
					$$->prev=NULL;
					$$->op=lnotOp;
					$$->value=0;
					$$->name=NULL;
				};
				
expression:
				expression '<' expression
				{
					$$=malloc(sizeof(operation));
					$$->left=$1;
					$$->right=$3;
					$$->next=NULL;
					$$->prev=NULL;
					$$->op=ltOp;
					$$->value=0;
					$$->name=NULL;
				};
				
expression:
				expression '>' expression
				{
					$$=malloc(sizeof(operation));
					$$->left=$1;
					$$->right=$3;
					$$->next=NULL;
					$$->prev=NULL;
					$$->op=gtOp;
					$$->value=0;
					$$->name=NULL;
				};
				
expression:
				expression '=' expression
				{
					$$=malloc(sizeof(operation));
					$$->left=$1;
					$$->right=$3;
					$$->next=NULL;
					$$->prev=NULL;
					$$->op=equOp;
					$$->value=0;
					$$->name=NULL;
				};

expression:
				expression SHL
				{
					$$=malloc(sizeof(operation));
					$$->left=NULL;
					$$->right=$1;
					$$->next=NULL;
					$$->prev=NULL;
					$$->op=shlOp;
					$$->value=0;
					$$->name=NULL;
				};
				
expression:
				expression SHR
				{
					$$=malloc(sizeof(operation));
					$$->left=NULL;
					$$->right=$1;
					$$->next=NULL;
					$$->prev=NULL;
					$$->op=shrOp;
					$$->value=0;
					$$->name=NULL;
				};
				
expression:
				expression BAND expression
				{
					$$=malloc(sizeof(operation));
					$$->left=$1;
					$$->right=$3;
					$$->next=NULL;
					$$->prev=NULL;
					$$->op=bandOp;
					$$->value=0;
					$$->name=NULL;
				};
				
expression:
				expression BOR expression
				{
					$$=malloc(sizeof(operation));
					$$->left=$1;
					$$->right=$3;
					$$->next=NULL;
					$$->prev=NULL;
					$$->op=borOp;
					$$->value=0;
					$$->name=NULL;
				};
				
expression:
				'~' expression
				{
					$$=malloc(sizeof(operation));
					$$->left=NULL;
					$$->right=$2;
					$$->next=NULL;
					$$->prev=NULL;
					$$->op=bnotOp;
					$$->value=0;
					$$->name=NULL;
				};
				
expression:
				expression '^' expression
				{
					$$=malloc(sizeof(operation));
					$$->left=$1;
					$$->right=$3;
					$$->next=NULL;
					$$->prev=NULL;
					$$->op=bxorOp;
					$$->value=0;
					$$->name=NULL;
				};
				

expression:
				expression '+' expression
				{
					$$=malloc(sizeof(operation));
					$$->left=$1;
					$$->right=$3;
					$$->next=NULL;
					$$->prev=NULL;
					$$->op=addOp;
					$$->value=0;
					$$->name=NULL;
				};
				
expression:
				expression '-' expression
				{
					$$=malloc(sizeof(operation));
					$$->left=$1;
					$$->right=$3;
					$$->next=NULL;
					$$->prev=NULL;
					$$->op=subOp;
					$$->value=0;
					$$->name=NULL;
				};
				
expression:
				expression '*' expression
				{
					$$=malloc(sizeof(operation));
					$$->left=$1;
					$$->right=$3;
					$$->next=NULL;
					$$->prev=NULL;
					$$->op=mulOp;
					$$->value=0;
					$$->name=NULL;
				};
				
expression:
				expression '/' expression
				{
					$$=malloc(sizeof(operation));
					$$->left=$1;
					$$->right=$3;
					$$->next=NULL;
					$$->prev=NULL;
					$$->op=divOp;
					$$->value=0;
					$$->name=NULL;
				};
				
expression:
				NAME INCOP
				{
					$$=malloc(sizeof(operation));
					$$->left=NULL;
					$$->right=NULL;
					$$->next=NULL;
					$$->prev=NULL;
					$$->op=incOp;
					$$->value=0;
					$$->name=malloc(NAMESIZE);
					strcpy($$->name,$1);
				};
				
expression:
				NAME DECOP
				{
					$$=malloc(sizeof(operation));
					$$->left=NULL;
					$$->right=NULL;
					$$->next=NULL;
					$$->prev=NULL;
					$$->op=decOp;
					$$->value=0;
					$$->name=malloc(NAMESIZE);
					strcpy($$->name,$1);
				};
				
statement:
				ASMBLOCK
				{
					$$=malloc(sizeof(statement));
					$$->upper=NULL;
					$$->lower=NULL;
					$$->next=NULL;
					$$->prev=NULL;
					$$->op=NULL;
					$$->construct=asmStatement;
					$$->var=NULL;
					$$->text=malloc(strlen($1)+1);
					strcpy($$->text,$1);
				};

statement:
				expression
				{
					$$=malloc(sizeof(statement));
					$$->upper=NULL;
					$$->lower=NULL;
					$$->next=NULL;
					$$->prev=NULL;
					$$->op=$1;
					$$->construct=exprStatement;
					$$->var=NULL;
					$$->text=NULL;
				};

statement:
				IF expression '{' statements '}'
				{
					$$=malloc(sizeof(statement));
					$$->upper=$4;
					$$->lower=NULL;
					$$->next=NULL;
					$$->prev=NULL;
					$$->op=$2;
					$$->construct=ifStatement;
					$$->var=NULL;
					$$->text=NULL;
				};
				
statement:
				IF expression '{' statements '}' '{' statements '}'
				{
					$$=malloc(sizeof(statement));
					$$->upper=$4;
					$$->lower=$7;
					$$->next=NULL;
					$$->prev=NULL;
					$$->op=$2;
					$$->construct=ifStatement;
					$$->var=NULL;
					$$->text=NULL;
				};
				
statement:
				WHILE expression '{' statements '}'
				{
					$$=malloc(sizeof(statement));
					$$->upper=$4;
					$$->lower=NULL;
					$$->next=NULL;
					$$->prev=NULL;
					$$->op=$2;
					$$->construct=whileStatement;
					$$->var=NULL;
					$$->text=NULL;
				};
				
statement:
				REPEAT '{' statements '}' WHILE expression
				{
					$$=malloc(sizeof(statement));
					$$->upper=$3;
					$$->lower=NULL;
					$$->next=NULL;
					$$->prev=NULL;
					$$->op=$6;
					$$->construct=repeatStatement;
					$$->var=NULL;
					$$->text=NULL;
				};
				
statement:
				RETURN expression
				{
					$$=malloc(sizeof(statement));
					$$->upper=NULL;
					$$->lower=NULL;
					$$->next=NULL;
					$$->prev=NULL;
					$$->op=$2;
					$$->construct=returnStatement;
					$$->var=NULL;
					$$->text=NULL;
				};

				
statement:
				BYTE NAME
				{
					$$=malloc(sizeof(statement));
					$$->upper=NULL;
					$$->lower=NULL;
					$$->next=NULL;
					$$->prev=NULL;
					$$->op=NULL;
					$$->construct=declarationStatement;
					$$->var=malloc(sizeof(variable));
					$$->var->name=malloc(NAMESIZE);
					strcpy($$->var->name,$2);
					$$->var->type=byte;
					$$->var->size=1;
					$$->text=NULL;
				};

%%
