/*
 *  mccCore.h
 *  mcc
 *
 *  Created by Britt Durbrow on 7/4/06.
 *  Copyright 2006 Rattlesnake Hill Softworks. All rights reserved.
 *
 */

#define NAMESIZE 32
#define STRINGBUFFERSIZE 65536


struct variable
{
	struct variable *next;
	struct variable *prev;
	char *name;
	unsigned int type;
	unsigned int size;
};

typedef struct variable variable;

enum
{
	byte=1,
	word,
	quadword
};

struct operation
{
	struct operation *left;
	struct operation *right;
	struct operation *next;
	struct operation *prev;
	unsigned int op;
	unsigned int value;
	char *name;
};

typedef struct operation operation;

enum
{
	varRefOp=0,
	varAddrOp,
	constOp,
	addOp,
	subOp,
	mulOp,
	divOp,
	gtOp,
	ltOp,
	equOp,
	landOp,
	lorOp,
	lnotOp,
	assignOp,
	bandOp,
	borOp,
	bnotOp,
	bxorOp,
	shlOp,
	shrOp,
	incOp,
	decOp,
	callOp
};

struct statement
{
	struct statement *next;
	struct statement *prev;
	struct statement *upper;
	struct statement *lower;
	operation *op;
	unsigned int construct;
	variable *var;
	char *text;
};

typedef struct statement statement;

enum
{
	asmStatement,
	exprStatement,
	ifStatement,
	whileStatement,
	repeatStatement,
	returnStatement,
	declarationStatement
};

struct function
{
	struct function *next;
	struct function *prev;
	struct statement *statements;
	struct variable *parameters;
	char *name;
	int kind;
};

typedef struct function function;

enum
{
	functionKind=0,
	vdxKind,
	initKind
};

struct program
{
	variable *globals;
	function *functions;
};

typedef struct program program;

extern program *parsedProgram;